@@ -12,21 +12,26 @@ local httpstream_new = require "net.http.parser".new;
 local util_http = require "util.http";
 local events = require "util.events";
 local verify_identity = require"util.x509".verify_identity;
+local promise = require "util.promise";
+local http_errors = require "net.http.errors";
+
+local basic_resolver = require "net.resolvers.basic";
+local connect = require "net.connect".connect;
 
 local ssl_available = pcall(require, "ssl");
 
-local server = require "net.server"
-
 local t_insert, t_concat = table.insert, table.concat;
 local pairs = pairs;
-local tonumber, tostring, xpcall, traceback =
-      tonumber, tostring, xpcall, debug.traceback;
+local tonumber, tostring, traceback =
+      tonumber, tostring, debug.traceback;
+local os_time = os.time;
+local xpcall = require "util.xpcall".xpcall;
 local error = error
-local setmetatable = setmetatable;
 
 local log = require "util.logger".init("http");
 
 local _ENV = nil;
+-- luacheck: std none
 
 local requests = {}; -- Open requests
 
@@ -34,9 +39,106 @@ local function make_id(req) return (tostring(req):match("%x+$")); end
 
 local listener = { default_port = 80, default_mode = "*a" };
 
+-- Request-related helper functions
+local function handleerr(err) log("error", "Traceback[http]: %s", traceback(tostring(err), 2)); return err; end
+local function log_if_failed(req, ret, ...)
+	if not ret then
+		log("error", "Request '%s': error in callback: %s", req.id, (...));
+		if not req.suppress_errors then
+			error(...);
+		end
+	end
+	return ...;
+end
+
+local function destroy_request(request)
+	local conn = request.conn;
+	if conn then
+		request.conn = nil;
+		conn:close()
+	end
+end
+
+local function cancel_request(request, reason)
+	if request.callback then
+		request.callback(reason or "cancelled", 0, request);
+		request.callback = nil;
+	end
+	if request.conn then
+		destroy_request(request);
+	end
+end
+
+local function request_reader(request, data, err)
+	if not request.parser then
+		local function error_cb(reason)
+			if request.callback then
+				request.callback(reason or "connection-closed", 0, request);
+				request.callback = nil;
+			end
+			destroy_request(request);
+		end
+
+		if not data then
+			error_cb(err);
+			return;
+		end
+
+		local finalize_sink;
+		local function success_cb(r)
+			if r.partial then
+				-- Request should be streamed
+				log("debug", "Request '%s': partial response (%s%s)",
+					request.id,
+					r.chunked and "chunked, " or "",
+					r.body_length and ("%d bytes"):format(r.body_length) or "unknown length"
+				);
+				if request.streaming_handler then
+					log("debug", "Request '%s': Streaming via handler");
+					r.body_sink, finalize_sink = request.streaming_handler(r);
+				end
+				return;
+			elseif finalize_sink then
+				log("debug", "Request '%s': Finalizing response stream");
+				finalize_sink(r);
+			end
+			if request.callback then
+				request.callback(r.body, r.code, r, request);
+				request.callback = nil;
+			end
+			destroy_request(request);
+		end
+		local function options_cb()
+			return request;
+		end
+		request.parser = httpstream_new(success_cb, error_cb, "client", options_cb);
+	end
+	request.parser:feed(data);
+end
+
+-- Connection listener callbacks
 function listener.onconnect(conn)
 	local req = requests[conn];
 
+	-- Initialize request object
+	req.write = function (...) return req.conn:write(...); end
+	local callback = req.callback;
+	req.callback = function (content, code, response, request)
+		do
+			local event = { http = req.http, url = req.url, request = req, response = response, content = content, code = code, callback = req.callback };
+			req.http.events.fire_event("response", event);
+			content, code, response = event.content, event.code, event.response;
+		end
+
+		log("debug", "Request '%s': Calling callback, status %s", req.id, code or "---");
+		return log_if_failed(req.id, xpcall(callback, handleerr, content, code, response, request));
+	end
+	req.reader = request_reader;
+	req.state = "status";
+	req.cancel = cancel_request;
+
+	requests[req.conn] = req;
+
 	-- Validate certificate
 	if not req.insecure and conn:ssl() then
 		local sock = conn:socket();
@@ -62,13 +164,11 @@ function listener.onconnect(conn)
 		t_insert(request_line, 4, "?"..req.query);
 	end
 
-	conn:write(t_concat(request_line));
-	local t = { [2] = ": ", [4] = "\r\n" };
 	for k, v in pairs(req.headers) do
-		t[1], t[3] = k, v;
-		conn:write(t_concat(t));
+		t_insert(request_line, k .. ": " .. v .. "\r\n");
 	end
-	conn:write("\r\n");
+	t_insert(request_line, "\r\n")
+	conn:write(t_concat(request_line));
 
 	if req.body then
 		conn:write(req.body);
@@ -79,7 +179,7 @@ function listener.onincoming(conn, data)
 	local request = requests[conn];
 
 	if not request then
-		log("warn", "Received response from connection %s with no request attached!", tostring(conn));
+		log("warn", "Received response from connection %s with no request attached!", conn);
 		return;
 	end
 
@@ -96,64 +196,32 @@ function listener.ondisconnect(conn, err)
 	requests[conn] = nil;
 end
 
+function listener.onattach(conn, req)
+	requests[conn] = req;
+	req.conn = conn;
+end
+
 function listener.ondetach(conn)
 	requests[conn] = nil;
 end
 
-local function destroy_request(request)
-	if request.conn then
-		request.conn = nil;
-		request.handler:close()
-	end
-end
-
-local function request_reader(request, data, err)
-	if not request.parser then
-		local function error_cb(reason)
-			if request.callback then
-				request.callback(reason or "connection-closed", 0, request);
-				request.callback = nil;
-			end
-			destroy_request(request);
-		end
-
-		if not data then
-			error_cb(err);
-			return;
-		end
-
-		local function success_cb(r)
-			if request.callback then
-				request.callback(r.body, r.code, r, request);
-				request.callback = nil;
-			end
-			destroy_request(request);
-		end
-		local function options_cb()
-			return request;
-		end
-		request.parser = httpstream_new(success_cb, error_cb, "client", options_cb);
-	end
-	request.parser:feed(data);
-end
-
-local function handleerr(err) log("error", "Traceback[http]: %s", traceback(tostring(err), 2)); end
-local function log_if_failed(id, ret, ...)
-	if not ret then
-		log("error", "Request '%s': error in callback: %s", id, tostring((...)));
-	end
-	return ...;
+function listener.onfail(req, reason)
+	req.http.events.fire_event("request-connection-error", { http = req.http, request = req, url = req.url, err = reason });
+	req.callback(reason or "connection failed", 0, req);
 end
 
 local function request(self, u, ex, callback)
 	local req = url.parse(u);
-	req.url = u;
 
 	if not (req and req.host) then
 		callback("invalid-url", 0, req);
 		return nil, "invalid-url";
 	end
 
+	req.url = u;
+	req.http = self;
+	req.time = os_time();
+
 	if not req.path then
 		req.path = "/";
 	end
@@ -166,7 +234,7 @@ local function request(self, u, ex, callback)
 		if ret then
 			return ret;
 		end
-		req, u, ex, callback = event.request, event.url, event.options, event.callback;
+		req, u, ex, req.callback = event.request, event.url, event.options, event.callback;
 	end
 
 	local method, headers, body;
@@ -204,6 +272,8 @@ local function request(self, u, ex, callback)
 			end
 		end
 		req.insecure = ex.insecure;
+		req.suppress_errors = ex.suppress_errors;
+		req.streaming_handler = ex.streaming_handler;
 	end
 
 	log("debug", "Making %s %s request '%s' to %s", req.scheme:upper(), method or "GET", req.id, (ex and ex.suppress_url and host_header) or u);
@@ -217,34 +287,17 @@ local function request(self, u, ex, callback)
 	end
 	local port_number = port and tonumber(port) or (using_https and 443 or 80);
 
+	local use_dane = self.options and self.options.use_dane;
 	local sslctx = false;
 	if using_https then
 		sslctx = ex and ex.sslctx or self.options and self.options.sslctx;
-	end
-
-	local handler, conn = server.addclient(host, port_number, listener, "*a", sslctx)
-	if not handler then
-		self.events.fire_event("request-connection-error", { http = self, request = req, url = u, err = conn });
-		callback(conn, 0, req);
-		return nil, conn;
-	end
-	req.handler, req.conn = handler, conn
-	req.write = function (...) return req.handler:write(...); end
-
-	req.callback = function (content, code, response, request)
-		do
-			local event = { http = self, url = u, request = req, response = response, content = content, code = code, callback = callback };
-			self.events.fire_event("response", event);
-			content, code, response = event.content, event.code, event.response;
+		if ex and ex.use_dane ~= nil then
+			use_dane = ex.use_dane;
 		end
-
-		log("debug", "Request '%s': Calling callback, status %s", req.id, code or "---");
-		return log_if_failed(req.id, xpcall(function () return callback(content, code, response, request) end, handleerr));
 	end
-	req.reader = request_reader;
-	req.state = "status";
 
-	requests[req.handler] = req;
+	local http_service = basic_resolver.new(host, port_number, "tcp", { servername = req.host; use_dane = use_dane });
+	connect(http_service, listener, { sslctx = sslctx }, req);
 
 	self.events.fire_event("request", { http = self, request = req, url = u });
 	return req;
@@ -253,9 +306,29 @@ end
 local function new(options)
 	local http = {
 		options = options;
-		request = request;
+		request = function (self, u, ex, callback)
+			if callback ~= nil then
+				return request(self, u, ex, callback);
+			else
+				return promise.new(function (resolve, reject)
+					request(self, u, ex, function (body, code, a, b)
+						if code == 0 then
+							reject(http_errors.new(body, { request = a }));
+						else
+							a.request = b;
+							resolve(a);
+						end
+					end);
+				end);
+			end
+		end;
 		new = options and function (new_options)
-			return new(setmetatable(new_options, { __index = options }));
+			local final_options = {};
+			for k, v in pairs(options) do final_options[k] = v; end
+			if new_options then
+				for k, v in pairs(new_options) do final_options[k] = v; end
+			end
+			return new(final_options);
 		end or new;
 		events = events.new();
 	};
@@ -263,7 +336,8 @@ local function new(options)
 end
 
 local default_http = new({
-	sslctx = { mode = "client", protocol = "sslv23", options = { "no_sslv2", "no_sslv3" } };
+	sslctx = { mode = "client", protocol = "sslv23", options = { "no_sslv2", "no_sslv3" }, alpn = "http/1.1", verify = "peer" };
+	suppress_errors = true;
 });
 
 return {
@@ -278,4 +352,8 @@ return {
 	urldecode = util_http.urldecode;
 	formencode = util_http.formencode;
 	formdecode = util_http.formdecode;
+	destroy_request = destroy_request;
+	features = {
+		sni = true;
+	};
 };
